from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Static, Header, Footer, Select, Input
from textual.containers import Container, Horizontal, Vertical
from enigmatui.data.enigma_config import EnigmaConfig
from enigmatui.screens.exit_configutation_without_saving_modal import ExitConfiguration
from enigmatui.screens.config_not_complete_modal import ConfigurationNotComplete
from enigmatui.screens.plugboard_content_invalid_modal import PlugboardContentInvalid
from enigmatui.widgets.plugboard_input import PlugboardInput

from enigmapython.Enigma import Enigma

# Enigma M3 components
from enigmapython.EnigmaM3 import EnigmaM3
from enigmapython.EnigmaM3RotorI import EnigmaM3RotorI
from enigmapython.EnigmaM3RotorII import EnigmaM3RotorII
from enigmapython.EnigmaM3RotorIII import EnigmaM3RotorIII
from enigmapython.EnigmaM3RotorIV import EnigmaM3RotorIV
from enigmapython.EnigmaM3RotorV import EnigmaM3RotorV
from enigmapython.EnigmaM3RotorVI import EnigmaM3RotorVI
from enigmapython.EnigmaM3RotorVII import EnigmaM3RotorVII
from enigmapython.EnigmaM3RotorVIII import EnigmaM3RotorVIII

# Enigma M4 components
from enigmapython.EnigmaM4 import EnigmaM4
from enigmapython.EnigmaM4RotorI import EnigmaM4RotorI
from enigmapython.EnigmaM4RotorII import EnigmaM4RotorII
from enigmapython.EnigmaM4RotorIII import EnigmaM4RotorIII
from enigmapython.EnigmaM4RotorIV import EnigmaM4RotorIV
from enigmapython.EnigmaM4RotorV import EnigmaM4RotorV
from enigmapython.EnigmaM4RotorVI import EnigmaM4RotorVI
from enigmapython.EnigmaM4RotorVII import EnigmaM4RotorVII
from enigmapython.EnigmaM4RotorVIII import EnigmaM4RotorVIII
from enigmapython.EnigmaM4RotorBeta import EnigmaM4RotorBeta
from enigmapython.EnigmaM4RotorGamma import EnigmaM4RotorGamma
from enigmapython.ReflectorUKWBThin import ReflectorUKWBThin
from enigmapython.ReflectorUKWCThin import ReflectorUKWCThin

from enigmapython.EtwPassthrough import EtwPassthrough
from enigmapython.SwappablePlugboard import SwappablePlugboard

from enigmapython.ReflectorUKWB import ReflectorUKWB
from enigmapython.ReflectorUKWC import ReflectorUKWC

class ConfigureScreen(Screen):

    enigma_config = EnigmaConfig()


    BINDINGS = [("ctrl+s", "save_and_exit", "Save and exit")
                #,("escape", "exit", "Exit")
               ]

    def action_exit(self):
        self.app.push_screen(ExitConfiguration())

    def action_save_and_exit(self):
        config_complete = True
        for select in self.query(".active"):
            if select.value  == Select.BLANK:
                config_complete = False
                break

        # Check if plugboard contains only letter pairs(and not single letters)
        if len(self.query_one("#plugboard", PlugboardInput).value.replace(" ", "")) % 2 != 0:
            self.app.push_screen(PlugboardContentInvalid())
         
        elif config_complete == True:
            etw = globals()[self.etw_type_select.value]()
            rotor0 =  globals()[self.rotor0_type_select.value](position=int(self.rotor0_position_select.value),ring=int(self.rotor0_ring_select.value))
            rotor1 =  globals()[self.rotor1_type_select.value](position=int(self.rotor1_position_select.value),ring=int(self.rotor1_ring_select.value))
            rotor2 =  globals()[self.rotor2_type_select.value](position=int(self.rotor2_position_select.value),ring=int(self.rotor2_ring_select.value))
            reflector = globals()[self.reflector_type_select.value]()
            plugboard = SwappablePlugboard()

            # Iterate through letter pairs
            pairs = self.plugboard_input.value.split()
            for i, pair in enumerate(pairs):
                plugboard.swap(pair[0],pair[1])

            if self.enigma_type_select.value == "EnigmaM3":
                self.enigma_config.set_configured_enigma(EnigmaM3(rotor1=rotor0, 
                                                     rotor2=rotor1, 
                                                     rotor3=rotor2, 
                                                     plugboard=plugboard,
                                                     etw=etw, 
                                                     reflector=reflector,
                                                     auto_increment_rotors=True
                                                     )
                                            )
            elif self.enigma_type_select.value == "EnigmaM4":
                rotor3 =  globals()[self.rotor3_type_select.value](position=int(self.rotor3_position_select.value),ring=int(self.rotor3_ring_select.value))
                self.enigma_config.set_configured_enigma(EnigmaM4(rotor1=rotor0, 
                                                     rotor2=rotor1, 
                                                     rotor3=rotor2,
                                                     rotor4=rotor3, 
                                                     plugboard=plugboard,
                                                     etw=etw, 
                                                     reflector=reflector,
                                                     auto_increment_rotors=True
                                                     )
                                            )
            self.app.pop_screen()
        else:
            self.app.push_screen(ConfigurationNotComplete())
            
            
            
        

        #self.app.pop_screen()

    def on_mount(self):
        self.plugboard_input = self.query_one("#plugboard", Input)
        self.enigma_type_select = self.query_one("#enigma_type", Select)
        self.etw_type_select = self.query_one("#etw_type", Select)
        self.rotor0_type_select = self.query_one("#rotor0_type", Select)
        self.rotor0_type_select = self.query_one("#rotor0_type", Select)
        self.rotor0_position_select = self.query_one("#rotor0_position", Select)
        self.rotor0_ring_select = self.query_one("#rotor0_ring", Select)
        self.rotor1_type_select = self.query_one("#rotor1_type", Select)
        self.rotor1_position_select = self.query_one("#rotor1_position", Select)
        self.rotor1_ring_select = self.query_one("#rotor1_ring", Select)
        self.rotor2_type_select = self.query_one("#rotor2_type", Select)
        self.rotor2_position_select = self.query_one("#rotor2_position", Select)
        self.rotor2_ring_select = self.query_one("#rotor2_ring", Select)
        self.rotor3_type_select = self.query_one("#rotor3_type", Select)
        self.rotor3_position_select = self.query_one("#rotor3_position", Select)
        self.rotor3_ring_select = self.query_one("#rotor3_ring", Select)
        self.reflector_type_select = self.query_one("#reflector_type", Select)


    def compose(self) -> ComposeResult:
        yield Header()
        yield Static("")
        yield Static("")
        yield Horizontal(
            Vertical(
                    Static("Select an Enigma machine model:"),
                    Select(options=[("Enigma M3", "EnigmaM3"),
                                    ("Enigma M4", "EnigmaM4"),
                                    # ("Enigma Z30 Mark I", "EnigmaZ30MarkI")
                                    ],id="enigma_type",
                                    allow_blank=False, 
                                    classes="active")
                )
            
        )
        yield Static("")
        yield Vertical(
                    Static("Plugboard:"),
                    PlugboardInput("", 
                          placeholder="Type letter pairs to scramble", 
                          id="plugboard"),
                    id="plugboard_vertical",
                    classes="invisible"
        )

        yield Vertical(
                    Static("Select ETW type:"),
                    Select(options=[],
                           id="etw_type",
                           classes="active"
                           )
                )
            
        yield Horizontal(
                Vertical(
                    Static("Select rotor 3 type:"),
                    Select( id="rotor3_type", options=[]),
                    id="rotor3_type_vertical",
                    classes="invisible"
                ),
                Vertical(
                    Static("Select rotor 2 type:"),
                    Select( id="rotor2_type", options=[], classes="active")
                ),
                Vertical(
                    Static("Select rotor 1 type:"),
                    Select( id="rotor1_type", options=[], classes="active")
                ),
                Vertical(
                    Static("Select rotor 0 type:"),
                    Select( id="rotor0_type", options=[], classes="active")
                ),
                id="rotors_types_line"
        )

        yield Horizontal(
                Vertical(
                    Static("Select rotor 3 position:"),
                    Select( id="rotor3_position", options=[]),
                    id="rotor3_position_vertical",
                    classes="invisible"
                ),
                Vertical(
                    Static("Select rotor 2 position:"),
                    Select( id="rotor2_position", options=[], classes="active")
                ),
                Vertical(
                    Static("Select rotor 1 position:"),
                    Select( id="rotor1_position", options=[], classes="active")
                ),
                Vertical(
                    Static("Select rotor 0 position:"),
                    Select( id="rotor0_position", options=[], classes="active")
                ),
                id="rotors_position_line"
        )

        yield Horizontal(
                Vertical(
                    Static("Select rotor 3 ring:"),
                    Select( id="rotor3_ring", options=[]),
                    id="rotor3_ring_vertical",
                    classes="invisible"
                ),
                Vertical(
                    Static("Select rotor 2 ring:"),
                    Select( id="rotor2_ring", options=[], classes="active")
                ),
                Vertical(
                    Static("Select rotor 1 ring:"),
                    Select( id="rotor1_ring", options=[], classes="active")
                ),
                Vertical(
                    Static("Select rotor 0 ring:"),
                    Select( id="rotor0_ring", options=[], classes="active")
                ),
                id="rotors_rings_line"
        )

        yield Horizontal(
            Vertical(
                    Static("Select a reflector:"),
                    Select(options=[],id="reflector_type",classes="active")
                )
        )

        yield Footer()

    async def on_input_changed(self, event: Input.Submitted) -> None:
        if event.input.id == "plugboard" :
            raw_text = event.value.replace(" ", "")  # Remove spaces for raw processing

            # Ensure that the last character is not duplicated
            current_text = event.value
            last_char = current_text[-1] if current_text else ""

            # If the last character is a duplicate, remove it
            if last_char and last_char in raw_text[:-1]:
                event.input.value = current_text[:-1]
                event.input.cursor_position = len(event.input.value)  # Keep cursor at the end
                return  # Skip further processing as we've already corrected

            # Format text to add spaces after every 2 characters
            spaced_text = " ".join(raw_text[i:i+2] for i in range(0, len(raw_text), 2))

            # Update the input value only if it has changed
            if event.input.value != spaced_text:
                event.input.value = spaced_text
                event.input.cursor_position = len(spaced_text)  # Move the cursor to the end
                    

    
    def on_select_changed(self, event: Select.Changed) -> None:
        if event.select.id == "enigma_type":
            

            self.enigma_config.unset_configured_enigma()

            self.query_one("#rotor3_type_vertical", Vertical).add_class("invisible")
            self.query_one("#rotor3_position_vertical", Vertical).add_class("invisible")
            self.query_one("#rotor3_ring_vertical", Vertical).add_class("invisible")
            self.query_one("#plugboard_vertical", Vertical).add_class("invisible")

            self.rotor3_type_select.remove_class("active")
            self.rotor3_position_select.remove_class("active")
            self.rotor3_ring_select.remove_class("active")
            self.plugboard_input.remove_class("active")
            
            if event.value == "EnigmaM3":

               

                self.etw_type_select.set_options([("Passthrough\t({})".format(EtwPassthrough.wiring).expandtabs(4), "EtwPassthrough")]) 
                self.etw_type_select.value="EtwPassthrough"

                self.query_one("#plugboard_vertical", Vertical).remove_class("invisible")
                self.plugboard_input.add_class("active")

                m3_rotors_options = [
                    ("I\t\t({})".format(EnigmaM3RotorI.wiring).expandtabs(4), "EnigmaM3RotorI"), 
                    ("II\t\t({})".format(EnigmaM3RotorII.wiring).expandtabs(4), "EnigmaM3RotorII"), 
                    ("III\t\t({})".format(EnigmaM3RotorIII.wiring).expandtabs(4), "EnigmaM3RotorIII"),
                    ("IV\t\t({})".format(EnigmaM3RotorIV.wiring).expandtabs(4), "EnigmaM3RotorIV"), 
                    ("V\t\t({})".format(EnigmaM3RotorV.wiring).expandtabs(4), "EnigmaM3RotorV"), 
                    ("VI\t\t({})".format(EnigmaM3RotorVI.wiring).expandtabs(4), "EnigmaM3RotorVI"),
                    ("VII\t\t({})".format(EnigmaM3RotorVII.wiring).expandtabs(4), "EnigmaM3RotorVII"), 
                    ("VIII\t({})".format(EnigmaM3RotorVIII.wiring).expandtabs(4), "EnigmaM3RotorVIII")]
                
                self.rotor0_type_select.set_options(m3_rotors_options)
                self.rotor0_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor0_position_select.value = "0"
                self.rotor0_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor0_ring_select.value = "0"

                self.rotor1_type_select.set_options(m3_rotors_options)
                self.rotor1_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor1_position_select.value = "0"
                self.rotor1_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor1_ring_select.value = "0"

                self.rotor2_type_select.set_options(m3_rotors_options)
                self.rotor2_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor2_position_select.value = "0"
                self.rotor2_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor2_ring_select.value = "0"

                self.reflector_type_select.set_options([
                    ("UKW-B\t({})".format(ReflectorUKWB.wiring).expandtabs(4), "ReflectorUKWB"), 
                    ("UKW-C\t({})".format(ReflectorUKWC.wiring).expandtabs(4), "ReflectorUKWC")])
                self.reflector_type_select.value = "ReflectorUKWB"
                self.reflector_type_select.allow_blank=False


            elif event.value == "EnigmaM4":
                self.query_one("#rotor3_type_vertical", Vertical).remove_class("invisible")
                self.query_one("#rotor3_position_vertical", Vertical).remove_class("invisible")
                self.query_one("#rotor3_ring_vertical", Vertical).remove_class("invisible")
                self.query_one("#plugboard_vertical", Vertical).remove_class("invisible")

                self.rotor3_type_select.add_class("active")
                self.rotor3_position_select.add_class("active")
                self.rotor3_ring_select.add_class("active")
                self.plugboard_input.add_class("active")

                self.etw_type_select.set_options([("Passthrough ({})".format(EtwPassthrough.wiring), "EtwPassthrough")]) 
                self.etw_type_select.value="EtwPassthrough"

                m4_rotors_options = [
                    ("I\t\t({})".format(EnigmaM4RotorI.wiring).expandtabs(4), "EnigmaM4RotorI"), 
                    ("II\t\t({})".format(EnigmaM4RotorII.wiring).expandtabs(4), "EnigmaM4RotorII"), 
                    ("III\t\t({})".format(EnigmaM4RotorIII.wiring).expandtabs(4), "EnigmaM4RotorIII"),
                    ("IV\t\t({})".format(EnigmaM4RotorIV.wiring).expandtabs(4), "EnigmaM4RotorIV"), 
                    ("V\t\t({})".format(EnigmaM4RotorV.wiring).expandtabs(4), "EnigmaM4RotorV"), 
                    ("VI\t\t({})".format(EnigmaM4RotorVI.wiring).expandtabs(4), "EnigmaM4RotorVI"),
                    ("VII\t\t({})".format(EnigmaM4RotorVII.wiring).expandtabs(4), "EnigmaM4RotorVII"), 
                    ("VIII\t\t({})".format(EnigmaM4RotorVIII.wiring).expandtabs(4), "EnigmaM4RotorVIII")]
                
                self.rotor0_type_select.set_options(m4_rotors_options)
                self.rotor0_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor0_position_select.value = "0"
                self.rotor0_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor0_ring_select.value = "0"

                self.rotor1_type_select.set_options(m4_rotors_options)
                self.rotor1_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor1_position_select.value = "0"
                self.rotor1_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor1_ring_select.value = "0"

                self.rotor2_type_select.set_options(m4_rotors_options)
                self.rotor2_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor2_position_select.value = "0"
                self.rotor2_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor2_ring_select.value = "0"

                self.rotor3_type_select.set_options([
                    ("Beta\t({})".format(EnigmaM4RotorBeta.wiring).expandtabs(4), "EnigmaM4RotorBeta"),
                    ("Gamma\t({})".format(EnigmaM4RotorGamma.wiring).expandtabs(4), "EnigmaM4RotorGamma")])
                self.rotor3_position_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor3_position_select.value = "0"
                self.rotor3_ring_select.set_options([(str(i), str(i)) for i in range(26)])
                self.rotor3_ring_select.value = "0"

                self.reflector_type_select.set_options([
                    ("UKW-B Thin\t({})".format(ReflectorUKWBThin.wiring).expandtabs(4), "ReflectorUKWBThin"), 
                    ("UKW-C Thin\t({})".format(ReflectorUKWCThin.wiring).expandtabs(4), "ReflectorUKWCThin")])
                self.reflector_type_select.value = "ReflectorUKWBThin"
                self.reflector_type_select.allow_blank=False
            elif event.value == "EnigmaZ30MarkI":
                self.query_one("#plugboard_vertical", Vertical).add_class("invisible")
                self.plugboard_input.remove_class("active")
            else:
                self.reset_form_options()
       
    def reset_form_options(self):
        self.plugboard_input.clear()
        self.enigma_type_select.clear()
        self.etw_type_select.set_options([])
        self.rotor0_type_select.set_options([])
        self.rotor0_position_select.set_options([])
        self.rotor0_ring_select.set_options([])
        self.rotor1_type_select.set_options([])
        self.rotor1_position_select.set_options([])
        self.rotor1_ring_select.set_options([])
        self.rotor2_type_select.set_options([])
        self.rotor2_position_select.set_options([])
        self.rotor2_ring_select.set_options([])
        self.rotor3_type_select.set_options([])
        self.rotor3_position_select.set_options([])
        self.rotor3_ring_select.set_options([])
        self.reflector_type_select.set_options([])
            
                


from textual.widgets import Input
from textual import events
from enigmatui.data.enigma_config import EnigmaConfig

class PlugboardInput(Input):

    async def on_input_changed(self, event: Input.Changed) -> None:
        event.value = event.value.lower()  # Convert input to lowercase
        """Sanitize input to allow only letters."""
        sanitized_value = "".join(c for c in event.value if c.isalpha() or c.isspace() or event.key == "backspace")
        if sanitized_value != event.value:
            event.input.value = sanitized_value  # Replace with sanitized input, lowercased


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _on_key(self,  event: events.Key):
        if event.key == "backspace" or event.key == "tab" or event.key == "shift+tab" or (event.key == "ctrl+s"):
            super()._on_key(event)
        elif (event.key == "left" or event.key == "space" or event.key == "ctrl-v") or (not event.is_printable):
            event.stop()
            event.prevent_default()
        else:
           super()._on_key(event)

    def on_focus(self, event: events.Focus) -> None:
        """Handle focus event to prevent full text selection."""
        # Use a small delay to ensure selection adjustments happen after focus logic
        self.call_later(self.clear_selection)
    
    def clear_selection(self) -> None:
        """Move cursor to the end and clear any selection."""
        self.cursor_position = len(self.value)  # Place the cursor at the end
        self.selection_range = None  # Clear selection
            


from textual.app import App, ComposeResult
from textual.containers import Vertical, Horizontal
from textual.widgets import Header, Footer, Button, Select, Static, Header
from textual.events import Key
from enigmatui.screens.splash import SplashScreen
from enigmatui.screens.main import MainScreen
from enigmatui.screens.configure import ConfigureScreen
from enigmatui.screens.encrypt import EncryptScreen
from enigmatui import __version__ as app_version


class EnigmaApp(App[str]):
   
    CSS_PATH = "css/styles.css"
    TITLE = "Enigma TUI v{}".format(app_version)
    
    SUB_TITLE = "A Terminal User Interface for Enigma machines by Denis Maggiorotto"

    def compose(self) -> ComposeResult:
        yield Header()

    def on_mount(self):
        self.install_screen(SplashScreen(), name="splash")
        self.install_screen(MainScreen(), name="main")
        self.push_screen("main") 
        self.push_screen("splash") 

   
        

    



class Observable:

    observers = None

    def __init__(self):
        self.observers = []

    def add_observer(self, observer):
        if observer not in self.observers:
            self.observers.append(observer)

    def remove_observer(self, observer):
        if observer in self.observers:
            self.observers.remove(observer)

    def notify_observers(self, *args, **kwargs):
        for observer in self.observers:
            observer.update(self, *args, **kwargs)


from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Static, Header, Footer, TextArea, Button
from textual.events import Paste
from textual.containers import Container, Horizontal, Vertical

from enigmatui.utility.observer import Observer
from enigmatui.widgets.undeletable_textarea import UndeletableTextArea
from enigmatui.data.enigma_config import EnigmaConfig
from enigmapython.XRay import XRay
import pyperclip

import re

class EncryptScreen(Screen,Observer):

    _prev_cleartext_area_text = ""

    BINDINGS = [("ctrl+r", "reset", "Reset Enigma"),
                ("escape", "back", "Back")]
    enigma_config = EnigmaConfig()

    def action_reset(self):
        self.enigma_config.reset_enigma()
        self.query_one("#cleartext", TextArea).clear()
        self.query_one("#ciphertext", TextArea).clear()
        self.update(None,None,None)
        self._prev_cleartext_area_text = ""

    def action_back(self):
        self.app.pop_screen()    

    def compose(self) -> ComposeResult:
        yield Header()
        yield Horizontal(
            Vertical(
                Static("", id="enigma-diagram"),
                id="enigma-diagram-vertical"
            ),
            Vertical(
                Static("", id="enigma-wirings"),
                id="enigma-wirings-vertical"
            ),
            id="enigma-diagram-wirings-horizontal"
        )
        yield Vertical(
            Horizontal(
                Static("Cleartext:", id="cleartext-label")
            ),
            Horizontal(
                UndeletableTextArea(id="cleartext"),
                Button("Copy", id="cleartext_copy"),
                Button("Paste", id="cleartext_paste")
            ),
            Static(""),
            Static(""),
            Horizontal(
                Static("Ciphertext:", id="ciphertext-label")
            ),
            Horizontal(
                UndeletableTextArea(id="ciphertext", read_only=True),
                Button("Copy", id="ciphertext_copy")
            )
        )
        
        yield Footer()

    def on_text_area_changed(self, event: TextArea.Changed) -> None:
        if event.text_area.id == "cleartext" and event.text_area.text != 'Type your cleartext here...' and event.text_area.text != "":
            # Save cleartext area (after the update)
            cleartext_area = event.text_area
            # Query the ciphertext area to update
            ciphertext_area = self.query_one("#ciphertext", TextArea)
            # Get the current cleartext area text (yet to be cleared)
            current_text = event.text_area.text
            # Get the previous cleartext area text (as it was before the update)
            prev_text = self._prev_cleartext_area_text
            # Remove from the cleartext area any character not in the alphabet list
            cleartext_area.text = re.sub(f"[^{''.join(self.enigma_config.enigma.alphabet_list)}]", "", current_text)
            # Update the _prev_cleartext_area_text with the update and cleared text
            self._prev_cleartext_area_text = cleartext_area.text
            # Set the cursor position to the end of the cleartext area
            cleartext_area.cursor_location = (len(cleartext_area.text.splitlines()) - 1, len(cleartext_area.text.splitlines()[-1]))
            # Append to the ciphertext area the just recently typed text (delta)
            ciphertext_area.text += self.enigma_config.enigma.input_string(cleartext_area.text[len(prev_text):])
            # Let the observers know that the enigma machine state has changed
            self.update(self.enigma_config, None, None)

        
    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "cleartext_copy":
            cleartext_area = self.query_one("#cleartext", TextArea)
            pyperclip.copy(cleartext_area.text)
        elif event.button.id == "cleartext_paste":
            paste_text =  pyperclip.paste()
            cleaned = re.sub(f"[^{''.join(self.enigma_config.enigma.alphabet_list)}]", "", paste_text)
            cleartext_area = self.query_one("#cleartext", TextArea)
            cleartext_area.text  += cleaned
        elif event.button.id == "ciphertext_copy":
            ciphertext_area = self.query_one("#ciphertext", TextArea)
            pyperclip.copy(ciphertext_area.text)


    def on_mount(self):
       self.enigma_config.add_observer(self)
       self.query_one("#enigma-diagram",Static).update(XRay.render_enigma_xray(self.enigma_config.enigma))
       self.query_one("#enigma-wirings",Static).update("Plugboard ({}) wiring: \n{}\n\n".format(type(self.enigma_config.enigma.plugboard).__name__,self.enigma_config.enigma.plugboard)+"ETW ({}) wiring: \n{}\n\n".format(type(self.enigma_config.enigma.etw).__name__,self.enigma_config.enigma.etw)+"\n".join(["Rotor {} ({}) wiring:\n{}\n".format(i,type(self.enigma_config.enigma.rotors[i]).__name__,self.enigma_config.enigma.rotors[i]) for i in range(len(self.enigma_config.enigma.rotors))])+"\nReflector ({}) wiring:\n{}\n".format(type(self.enigma_config.enigma.reflector).__name__,self.enigma_config.enigma.reflector))


    def update(self, observable, *args, **kwargs):
        self.query_one("#enigma-diagram",Static).update(XRay.render_enigma_xray(self.enigma_config.enigma))
        self.query_one("#enigma-wirings",Static).update("Plugboard ({}) wiring: \n{}\n\n".format(type(self.enigma_config.enigma.plugboard).__name__,self.enigma_config.enigma.plugboard)+"ETW ({}) wiring: \n{}\n\n".format(type(self.enigma_config.enigma.etw).__name__,self.enigma_config.enigma.etw)+"\n".join(["Rotor {} ({}) wiring:\n{}\n".format(i,type(self.enigma_config.enigma.rotors[i]).__name__,self.enigma_config.enigma.rotors[i]) for i in range(len(self.enigma_config.enigma.rotors))])+"\nReflector ({}) wiring:\n{}\n".format(type(self.enigma_config.enigma.reflector).__name__,self.enigma_config.enigma.reflector))
    
    #def on_screen_resume(self) -> None:
    #    self.query_one("#ciphertext", TextArea).clear()
    #    self.query_one("#cleartext", TextArea).clear()
            


from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Button, Label
from textual import on
from textual.containers import Container, Horizontal

class ConfigurationNotComplete(ModalScreen):
    """A modal exit screen."""

    def compose(self) -> ComposeResult:
        with Container():
            yield Label("Enigma configuration is not valid, please fill (all) the empty field(s) before saving!")
            with Horizontal():
                yield Button("ok", id="ok")

    @on(Button.Pressed, "#ok")
    def back_to_app(self) -> None:
        self.app.pop_screen()


from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.containers import Horizontal, Vertical
from textual.widgets import Static, Header, Footer, Button
from enigmatui.screens.configure import ConfigureScreen
from enigmatui.screens.encrypt import EncryptScreen
from enigmatui.data.enigma_config   import EnigmaConfig

from enigmatui.utility.observer import Observer
from enigmatui.screens.config_null_modal import ConfigurationNull

class MainScreen(Screen,Observer):

    enigma_config = EnigmaConfig()

    BINDINGS = [("c", "go_to_configure", "Configure"),("e", "go_to_encrypt", "De/Encrypt")]

    

    def action_go_to_configure(self):
        self.app.push_screen("configure") 

    def action_go_to_encrypt(self):
        if not self.enigma_config.enigma:
            self.app.push_screen(ConfigurationNull())
        else:
            self.app.push_screen("encrypt")

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static("What do you want to do?", id="main-text")
        yield Footer()

    def on_mount(self):
        self.enigma_config.add_observer(self)
        self.app.install_screen(ConfigureScreen(), name="configure")
        self.app.install_screen(EncryptScreen(), name="encrypt")

    def update(self, observable, *args, **kwargs):    
        None

   


from textual.widgets import TextArea
from textual import events
from enigmatui.data.enigma_config import EnigmaConfig

import re

class UndeletableTextArea(TextArea):

    enigma_config = EnigmaConfig()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _on_key(self,  event: events.Key):
        #if event.key == "backspace" or event.key == "enter" :
        if (event.key == "enter" or event.key == "space" or event.key == "backspace" or event.key == "ctrl+d") or (event.is_printable and event.key not in self.enigma_config.enigma.alphabet_list):
            event.stop()
            event.prevent_default()
        else:
           super()._on_key(event)



from textual.app import App, ComposeResult
from textual.screen import Screen
from textual.widgets import Static, Header, Footer
import asyncio

SPLASH_TEXT = """Enigma TUI is a Terminal User Interface for Enigma machines, 
allowing you to simulate different Enigma machine models from the terminal."""



ENIGMA_DIAG = """
                UKW        Rotor       Rotor       Rotor        ETW      PLUGBOARD
                             2           1           0      
              +-----+     +-----+     +-----+     +-----+     +-----+     +-----+
              |     |     |     |     |     |     |     |     |     |     |     |
              |  +--|--<--|-----|--<--|-----|--<--|-----|--<--|-----|--<--|-----|----< d <-- Key
              |  |  |  w  |     |  n  |     |  l  |     |  d  |     |  d  |     |
              |  |  |     |     |     |     |     |     |     |     |     |     |
              |  |  |     |     |     |     |     |     |     |     |     |     |
              |  |  |  v  |     |  i  |     |  v  |     |  n  |     |  m  |     |     
              |  +--|-->--|-----|-->--|-----|-->--|-----|-->--|-----|-->--|-----|----> n --> Lamp
              |     |     |     |     |     |     |     |     |     |     |     |
              +-----+     +-----+     +-----+     +-----+     +-----+     +-----+

        Pos.:                00          00          01    
        Ring:                00          00          00
"""
class SplashScreen(Screen):

    BINDINGS = [("enter", "app.pop_screen", "Go to app")]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Static(" Welcome! ", id="welcome")
        yield Static(SPLASH_TEXT, id="splash-text")
        yield Static(ENIGMA_DIAG, id="enigma-diagram")
        yield Static("On every screen, ⬇️  see in the bar at the bottom ⬇️  the possible options. Now press enter to continue.", id="any-key")
        yield Footer()
    
 



from enigmatui.utility.observable import Observable
import copy
class EnigmaConfig(Observable):
    _instance = None
   
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(EnigmaConfig, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if not self._initialized:
            super().__init__()  # Initialize the observer
            self.configured_enigma = None  # 
            self.enigma = None 
            self._initialized = True
    
    def set_configured_enigma(self, enigma):
        self.configured_enigma = enigma
        self.enigma = self.configured_enigma.clone()
        self.notify_observers(self,None,None)
    
    def unset_configured_enigma(self):
        self.configured_enigma = None
        self.enigma = None

    def reset_enigma(self):
        self.enigma = self.configured_enigma.clone()
        self.notify_observers(self,None,None)
    
    
   
   


from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Button, Label
from textual import on
from textual.containers import Container, Horizontal

class ExitConfiguration(ModalScreen):
    """A modal exit screen."""

    def compose(self) -> ComposeResult:
        with Container():
            yield Label("Are you sure you want to exit configuration screen without saving?")
            with Horizontal():
                yield Button("no", id="no")
                yield Button("yes", id="yes")
    
    @on(Button.Pressed, "#yes")
    def exit_app(self) -> None:
        self.app.pop_screen()
        self.app.pop_screen()

    @on(Button.Pressed, "#no")
    def back_to_app(self) -> None:
        self.app.pop_screen()


from setuptools import setup, find_packages
from pathlib import Path
import os
import re


this_directory = Path(__file__).parent
long_description = (this_directory/".."/".."/"README.md").read_text()

def read_version():
    with open(os.path.join("enigmatui", "__init__.py")) as f:
        match = re.search(r'__version__\s*=\s*[\'"]([^\'"]+)[\'"]', f.read())
        if match:
            return match.group(1)
        raise RuntimeError("Unable to find version string.")


setup(
    author="Denis Maggiorotto",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/denismaggior8/enigma-tui",
    name="enigmatui",
    version=read_version(),  # Use a function to read the version from __init__.py
    include_package_data=True,
    packages=find_packages(
        # All keyword arguments below are optional:
        where='.',  # '.' by default
    ),
    include_dirs= ["css"],
    package_data={'enigmatui': ['css/*.css', 'css/*.css']},
    entry_points={
        "console_scripts": [
            "enigmatui = enigmatui.__main__:main",  # 
        ],
    },
    install_requires=[
        "textual==1.0.0",
        "enigmapython==1.3.0"
    ],
    description="Enigma TUI is a Terminal User Interface for Enigma machines, allowing you to simulate different Enigma machine models from the terminal"
)